import ctypes as _ct
import math as _math

import numpy as _np
from typing import Iterable 
from numbers import Real

from .._ctypeslib import pydll as _pydll




def kurt(y:Iterable)->float:
	"""
	Computes excess kurtosis.
	"""
	n = len(y)

	assert n >= 4, "y must have at least 4 elements"
	assert isinstance(y, Iterable), "Iterable object expected"
	
	Arr = _np.asfarray(y)

	avg = _np.mean(Arr)
	stdev = _np.std(Arr, ddof=1)

	s = 0
	for Num in y:
		s += (Num - avg)**4

	Kurt = (s/n)/(stdev)**4 - 3

	return float(Kurt)



def mode(y:Iterable)->tuple[list[Real], int]:
	"""
	Computes mode

	Returns the most frequently occuring number(s) and 
	the number of times it occurs.
	"""
	n = len(y)

	assert n >= 3, "y must have at least 3 elements"
	assert isinstance(y, Iterable), "Iterable object expected"
	
	Arr = _np.asarray(y)

	values, counts = _np.unique(Arr, return_counts=True)
	if(len(values) == len(Arr)):
		return ([], 0)

	index = _np.argwhere(counts == _np.max(counts))

	return values[index].flatten().tolist(), int(_np.max(counts))




class rolling:

	"""
	A class to compute rolling window mean, std, var min, max...

	## Example:
	>> r = rolling(x, y, period=2) \n
	>> r.mean() \n
	>> r.var() 
	"""

     
	def __init__(
			self, 
			x:Iterable, 
			y:Iterable, 
			period:int = 2):
		assert isinstance(x, Iterable), "x must be of type Iterable"
		assert isinstance(y, Iterable), "y must be of type Iterable"
		assert isinstance(period, int), "period must be of type int"

		self._m_X, self._m_Windows = _pydll.c_stat_rolling(_ct.py_object(x), _ct.py_object(y), _ct.c_int(period))
		

	def mean(self)->list[Real]:
		"""computes mean"""
		return self.__compute(self.__mean)

	def median(self)->list[Real]:
		"""computes median"""
		return self.__compute(self.__median)
	
	def min(self)->list[Real]:
		"""computes minimum"""
		return self.__compute(min)
	
	def max(self)->list[Real]:
		"""computes maximum"""
		return self.__compute(max)
	

	def __compute(self, func)->list:
		"""
		Calls the parameter func on self._m_Windows
		Returns a 1D list containing real numbers
		"""
		return [func(v) for v in self._m_Windows]

	
	def __mean(self, lst:list):
		return sum(lst)/len(lst)

	def __median(self, arg:Iterable):
		arr = _np.asfarray(arg)
		return float(_np.median(arr))


def moveavg(x, y, period=2)->tuple[list[Real], list[Real]]:
	"""
	Computes the moving average, similar to rolling.mean()
	returns corresponding x and mean values
	"""
	c = rolling(x,y, period)
	return x[period-1:], c.mean()





def skew(y:Iterable)->float:
	"""
	Computes the skewness of a distribution
	"""
	n = len(y)

	assert n >= 3, "y must have at least 3 elements"
	assert isinstance(y, Iterable), "Iterable object expected"
	
	Arr = _np.asarray(y)
	
	avg, std_p = float(_np.mean(Arr)), float(_np.std(Arr))

	skew_p = float(_np.sum((Arr-avg)**3))
	skew_p /= float(n * std_p**3.0)

	return _math.sqrt(n * (n - 1)) * skew_p/(n-2)


import ctypes as _ct
import dataclasses as _dc
import numbers as _numbers
import types as _types

from ._ctypeslib import pydll as _pydll


@_dc.dataclass
class BracketResult:
	a:float
	b:float
	c:float
	fa:float
	fb:float
	fc:float


def bracket(
	f:_types.FunctionType, 
	xa:_numbers.Real = 0.0, 
	xb:_numbers.Real = 1.0, 
	grow_limit=110,
	maxiter=1000)->BracketResult:
	"""
	Bracket the minimum of a function.
	Returns three points that bracket the minimum of the function.

	## Inputs:
	f: A unary function
	xa, xb: Initial guesses, local minimum need not be contained within this interval.
	grow_limit: Maximum grow limit.
	maxiter: Maximum number of iterations

	## References:
	- Press WH, Teukolsky SA, Vetterling WT, Flannery BP (2007). 
	  Numerical Recipes The Art of Scientific Computing. Cambridge Uni Press.
	"""
	assert isinstance(f, _types.FunctionType), "f must be function"
	assert isinstance(xa, _numbers.Real), "xa must be real number"
	assert isinstance(xb, _numbers.Real), "xb must be real number"
	assert isinstance(grow_limit, _numbers.Real), "grow_limit must be real number"
	assert isinstance(maxiter, int), "maxiter must be int"

	assert maxiter>0, "maxiter>0 expected"

	d:dict =_pydll.c_optimize_bracket(_ct.py_object(f), 
						_ct.c_double(xa), 
						_ct.c_double(xb),
						_ct.c_double(grow_limit),
						_ct.c_uint32(maxiter))
	
	return BracketResult(a=d["a"], b=d["b"], c=d["c"], fa=d["fa"], fb=d["fb"], fc=d["fc"])



#-------------------------------------------------------------------

@_dc.dataclass
class GoldenResult:
	xopt:float
	err: float
	iter: int


def golden(
	f:_types.FunctionType, 
	xlow:_numbers.Real, 
	xhigh:_numbers.Real, 
	tol=1E-6,
	maxiter=1000)->GoldenResult:
	"""
	Finds the *local* minimum using golden section method

	## Inputs:
	f: A unary function
	xlow, xhigh: Initial guesses, local minimum need to be contained within this interval.
	tol: tolerance.
	maxiter: Maximum number of iterations

	## References:
	- Chapra SC, Canale RP (2013). Numerical Methods for Engineers, 7th Ed. McGraw Hill Education.
	- Press WH, Teukolsky SA, Vetterling WT, Flannery BP (2007). 
	  Numerical Recipes The Art of Scientific Computing. Cambridge University Press.
	"""
	assert isinstance(f, _types.FunctionType), "f must be function"
	assert isinstance(xlow, _numbers.Real), "xlow must be real number"
	assert isinstance(xhigh, _numbers.Real), "xhigh must be real number"
	assert isinstance(tol, _numbers.Real), "tol must be real number"
	assert isinstance(maxiter, int), "maxiter must be int"

	assert maxiter>0, "maxiter>0 expected"

	xopt, err, iter =_pydll.c_optimize_golden(_ct.py_object(f), 
						_ct.c_double(xlow), 
						_ct.c_double(xhigh),
						_ct.c_double(tol),
						_ct.c_uint32(maxiter))
	
	return GoldenResult(xopt=xopt, err=err, iter=iter)




#-----------------------------------------------------------------

@_dc.dataclass
class BrentResult:
	xopt:float
	fval: float
	iter: int


def brent(
	f:_types.FunctionType, 
	xlow:_numbers.Real, 
	xhigh:_numbers.Real, 
	maxiter=500)->BrentResult:
	"""
	Finds the minimum using Brent's method

	## Inputs:
	f: A unary function
	xlow, xhigh: Initial guesses, local minimum need to be contained within this interval.
	maxiter: Maximum number of iterations

	## References
	- https://www.boost.org/doc/libs/1_82_0/libs/math/doc/html/math_toolkit/brent_minima.html
	"""
	assert isinstance(f, _types.FunctionType), "f must be function"
	assert isinstance(xlow, _numbers.Real), "xlow must be real number"
	assert isinstance(xhigh, _numbers.Real), "xhigh must be real number"
	assert isinstance(maxiter, int), "maxiter must be int"

	assert maxiter>0, "maxiter>0 expected"

	xopt, fxopt, iter =_pydll.c_optimize_brent(_ct.py_object(f), 
						_ct.c_double(xlow), 
						_ct.c_double(xhigh),
						_ct.c_longlong(maxiter))
	
	return BrentResult(xopt=xopt, fval=fxopt, iter=iter)



#----------------------------------------------------------

@_dc.dataclass
class ParabolicResult:
	xopt:float
	err: float
	iter: int


def parabolic(
	f:_types.FunctionType, 
	xa:_numbers.Real, 
	xb:_numbers.Real, 
	xc:None | _numbers.Real = None,
	tol=1E-6,
	maxiter=1000)->ParabolicResult:
	"""
	Finds the minimum using quadratic interpolation algorithm

	## Inputs:
	f: A unary function
	xa, xb, xc: initial guesses
	tol: tolerance.
	maxiter: Maximum number of iterations

	## References: 
	- Chapra SC, Canale RP (2013). Numerical Methods for Engineers, 7th Ed. McGraw Hill Education.
	- Cheney W, Kincaid D (2007). Numerical Mathematics and Computing, 6th Ed. Brooks Cole 
	"""
	assert isinstance(f, _types.FunctionType), "f must be function"
	assert isinstance(xa, _numbers.Real), "xa must be real"
	assert isinstance(xb, _numbers.Real), "xb must be real"
	assert isinstance(xc, _numbers.Real | None), "xc must be None or real"
	assert isinstance(tol, _numbers.Real), "tol must be real"
	assert isinstance(maxiter, int), "maxiter must be int"

	assert maxiter>0, "maxiter>0 expected"

	assert abs(xa-xb)>tol, "xa and xb must be different"
	if isinstance(xc, _numbers.Real):
		assert abs(xc-xb), "xa != xb != xc expected"

	xopt, err, iter =_pydll.c_optimize_parabolic(_ct.py_object(f), 
						_ct.c_double(xa), 
						_ct.c_double(xb),
						_ct.py_object(xc),
						_ct.c_double(tol),
						_ct.c_uint32(maxiter))
	
	return ParabolicResult(xopt=xopt, err=err, iter=iter)



import math
from typing import Iterable

from ..roots import brentq
from ._charts import canvas, title, xlabel, xscale, ylabel, yscale
from .gdi import arrow, curve, line, rect, text, makegroup


def _Colebrook(f, Re, E_D):
	temp1 = E_D/3.7 + 2.51/(Re*math.sqrt(f)) 
	temp2 = 2.0* math.log10(temp1) 
	return 1/math.sqrt(f) + temp2


def _TurbulenceOnset(E_D:float):	
	"""
	compute friction factor
	A modified version of Colebrook's equation for large Re
	"""
	def _vonKarman(E_D:float):
		f = -2.0* math.log10(E_D/3.7) 
		return (1/f)**2
	
	"""
	compute Reynolds number
	Transactions of the ASME, 66 (8): 671-684 (Page: 676)
	"""
	def _Reynolds(f:float):
		return 1/math.sqrt(f)*200/E_D
	
	f = _vonKarman(E_D)
	Re = _Reynolds(f)
	return Re, f



def moody(
		Re:Iterable = [1E3, 1E7],
		friction:Iterable = [0.01, 0.09]):
	"""
	`Re:` Reynolds number range
	`friction:`  Friction factor range
	"""
	assert isinstance(Re, Iterable), "Re must be iterable."
	assert len(Re)==2, "Re must have length 2."

	assert isinstance(friction, Iterable), "friction must be iterable."
	assert len(friction)==2, "friction must have length 2."

	canvas(
		x=(float(Re[0]), float(Re[1])), 
		y=(float(friction[0]), float(friction[1])))
	xscale("log")
	yscale("log")

	"""
	laminar flow 0<Re<2300
	"""
	fD = lambda _Re: 64.0/_Re
	p1 = (1000, fD(1000))
	p2 = (2300, fD(2300))
	idline = line(p1=p1, p2=p2, lw=2, ls="--", ec="#FF0000", label="laminar")
	idtxt = text(xy=p2, label="Laminar Flow\nf = 64/Re", hanchor="c")
	makegroup(idline, [idtxt])

	"""
	transition region 2300<Re<4000
	btmleft: x (onset of transition), y(roughly 64/Re)
	rectangle's width = 4000-2300=1700, height=0.08-0.03
	"""
	btmleft = (2300, 0.03)
	idrect = rect(xy=btmleft, width=1700, height=0.05, hatch="solid", alpha=0.5, ls="--", fc="#808080", label="transition")
	idtext = text(xy=(3300, 0.06), label="Transition Region", rotation=-90, labelcolor="#A52A2A")
	makegroup(idrect, [idtext])

	"""
	Turbulent Region - Re>=4000
	"""
	Re_Turbulent = [4000, 6000, 9000, 12E3, 20E3, 30E3, 60E3, 120E3, 250E3, 1E6, 5E6]
	e_d = [5E-5, 1E-4, 5E-4, 1E-3,2.5E-3, 5E-3, 0.01, 0.02, 0.03, 0.04, 0.05]

	TurbulenceLine = []
	for _e_d in e_d:
		x, y = [], []
		TurbulenceLine.append(_TurbulenceOnset(_e_d))
		for _Re in Re_Turbulent:
			_cb = lambda f: _Colebrook(f, _Re, _e_d)
			friction, _ = brentq(_cb, a=0.001, b=0.08)
			x.append(_Re)
			y.append(friction)

		id1 = text(xy=(_Re, friction), label=str(_e_d), vanchor="c")
		idcurve = curve(x, y, lw=2, ec="#A52A2A", label=str(_e_d))
		makegroup(idcurve, [id1])


	"""
	fully rough turbulent flow line
	E/D=0.06 appended so that the dashed line will be slightly above E/D=0.05 line
	"""
	TurbulenceLine.append(_TurbulenceOnset(0.06))
	_turbulence = list(zip(*TurbulenceLine))
	idturbcurve = curve(x=_turbulence[0], y=_turbulence[1], lw=2, ls="--", label="complete turb")
	
	midpoint = int(len(TurbulenceLine)/2)
	xx, yy=_turbulence[0][midpoint], _turbulence[1][midpoint]
	_p1 = (15E3, 0.015)
	idturbarr = arrow(p1=_p1, p2=(xx,yy), lw=2, length=0.05)
	idturbtxt = text(xy=_p1, label="Complete Turbulence", hanchor="c")
	makegroup(idturbcurve, [idturbtxt, idturbarr])


	xlabel("Reynolds Number")
	ylabel("Friction Factor")
	title("Moody Diagram")



from ctypes import PyDLL, c_bool, c_char, c_char_p, c_double, c_int, py_object, c_uint8, c_uint32, c_longlong, c_ulonglong
import sys
from pathlib import Path as _Path

"""
Tested with: 3.10.6, 3.11.6, 3.12.0
"""
_DLLname = "scisuit_pybind" + str(sys.version_info.major) + str(sys.version_info.minor)


#__file__ is guaranteed to be an absolute path in Python 3.9+
__pt = _Path(__file__)
pydll = PyDLL(str(__pt.parents[0] / _DLLname))



__all__ = ['pydll']




"""
-------------------------------------------------------------------------------
------------------------  ENGINEERING --------------------------------------- 
"""

pydll.c_eng_psychrometry.argtypes = [py_object]
pydll.c_eng_psychrometry.restype = py_object




"""
-------------------------------------------------------------------------------
------------------------    ROOTS  --------------------------------------- 
"""

pydll.c_root_bisect.argtypes = [py_object, c_double, c_double, c_double, c_int, c_char_p, c_bool]
pydll.c_root_bisect.restype = py_object


pydll.c_root_itp.argtypes = [
					py_object, #f
					c_double, #a
					c_double, #b
					c_double, #k1
					c_double, #k2
					c_double, #TOL
					c_int] #maxiter
pydll.c_root_itp.restype = py_object


pydll.c_root_brentq.argtypes = [
					py_object, 
					c_double, 
					c_double, 
					c_double, 
					c_int]
pydll.c_root_brentq.restype = py_object


pydll.c_root_muller.argtypes = [
					py_object, 
					py_object, 
					py_object, 
					py_object, 
					py_object, 
					c_double, 
					c_int]
pydll.c_root_muller.restype = py_object


pydll.c_root_newton.argtypes = [
					py_object, #f
					c_double, #X0
					py_object, #X1
					py_object, #fprime
					py_object, #fprime2
					c_double, #tol
					c_int #maxiter
				]
pydll.c_root_newton.restype = py_object


pydll.c_root_ridder.argtypes = [
					py_object, 
					c_double, 
					c_double, 
					c_double, 
					c_int]
pydll.c_root_ridder.restype = py_object


pydll.c_root_toms748.argtypes = [
					py_object, 
					c_double, 
					c_double, 
					c_double, 
					c_int]
pydll.c_root_toms748.restype = py_object




"""
-------------------------------------------------------------------------------
------------------------  FITTING PACKAGE --------------------------------------- 
"""

pydll.c_fit_expfit.argtypes = [py_object, py_object, py_object]
pydll.c_fit_expfit.restype = py_object


pydll.c_fit_lagrange.argtypes = [py_object, py_object,c_double]
pydll.c_fit_lagrange.restype = py_object


pydll.c_fit_logfit.argtypes = [py_object, py_object]
pydll.c_fit_logfit.restype = py_object


pydll.c_fit_logistfit.argtypes = [py_object, py_object, py_object]
pydll.c_fit_logistfit.restype = py_object


pydll.c_fit_powfit.argtypes = [py_object, py_object]
pydll.c_fit_powfit.restype = py_object


pydll.c_fit_spline.argtypes = [py_object, py_object]
pydll.c_fit_spline.restype = py_object






"""
-------------------------------------------------------------------------------
------------------------  INTEGRATION --------------------------------------- 
"""
pydll.c_integ_simpson.argtypes = [py_object, py_object]
pydll.c_integ_simpson.restype = py_object

pydll.c_integ_romberg.argtypes = [py_object, c_double, c_double, c_double, c_int]
pydll.c_integ_romberg.restype = py_object

pydll.c_integ_fixed_quad.argtypes = [py_object, c_double, c_double, c_int]
pydll.c_integ_fixed_quad.restype = py_object



"""
-------------------------------------------------------------------------------
------------------------  OPTIMIZATION --------------------------------------- 
"""

pydll.c_optimize_bracket.argtypes = [
					py_object, #function
					c_double, #a
					c_double, #b
					c_double, #growlimit
					c_uint32] #maxiter
pydll.c_optimize_bracket.restype = py_object 


pydll.c_optimize_golden.argtypes = [
					py_object, #function
					c_double, #xlow
					c_double, #xhigh
					c_double, #tol
					c_uint32] #maxiter
pydll.c_optimize_golden.restype = py_object 


pydll.c_optimize_parabolic.argtypes = [
					py_object, #function
					c_double, #xa
					c_double, #xb
					py_object, #xc
					c_double, #tol
					c_uint32] #maxiter
pydll.c_optimize_parabolic.restype = py_object 


pydll.c_optimize_brent.argtypes = [
					py_object, #function
					c_double, #xlow
					c_double, #xhigh
					c_longlong] #maxiter
pydll.c_optimize_brent.restype = py_object 


"""
-------------------------------------------------------------------------------
------------------------  STATS --------------------------------------- 
"""

pydll.c_stat_dbeta.argtypes = [py_object,  c_double, c_double]
pydll.c_stat_dbeta.restype=py_object

pydll.c_stat_pbeta.argtypes = [py_object,  c_double, c_double]
pydll.c_stat_pbeta.restype=py_object

pydll.c_stat_qbeta.argtypes = [py_object,  c_double, c_double]
pydll.c_stat_qbeta.restype=py_object

#----

pydll.c_stat_dbinom.argtypes = [py_object, c_int, c_double]
pydll.c_stat_dbinom.restype=py_object

pydll.c_stat_pbinom.argtypes = [py_object, c_int, c_double]
pydll.c_stat_pbinom.restype=py_object

pydll.c_stat_qbinom.argtypes = [py_object, c_int, c_double]
pydll.c_stat_qbinom.restype=py_object

#----

pydll.c_stat_dnbinom.argtypes = [py_object, c_int, c_double]
pydll.c_stat_dnbinom.restype=py_object

pydll.c_stat_pnbinom.argtypes = [py_object, c_int, c_double]
pydll.c_stat_pnbinom.restype=py_object

pydll.c_stat_qnbinom.argtypes = [py_object, c_int, c_double]
pydll.c_stat_qnbinom.restype=py_object

#----

pydll.c_stat_dmultinom.argtypes = [py_object, c_int, py_object]
pydll.c_stat_dmultinom.restype=py_object

#----

pydll.c_stat_dchisq.argtypes = [py_object, c_int]
pydll.c_stat_dchisq.restype=py_object

pydll.c_stat_pchisq.argtypes = [py_object, c_int]
pydll.c_stat_pchisq.restype=py_object

pydll.c_stat_qchisq.argtypes = [py_object, c_int]
pydll.c_stat_qchisq.restype=py_object

#----

pydll.c_stat_dexp.argtypes = [py_object, c_double]
pydll.c_stat_dexp.restype=py_object

pydll.c_stat_pexp.argtypes = [py_object, c_double]
pydll.c_stat_pexp.restype=py_object

pydll.c_stat_qexp.argtypes = [py_object, c_double]
pydll.c_stat_qexp.restype=py_object

#----

pydll.c_stat_df.argtypes = [py_object, c_int, c_int]
pydll.c_stat_df.restype=py_object

pydll.c_stat_pf.argtypes = [py_object, c_int, c_int]
pydll.c_stat_pf.restype=py_object

pydll.c_stat_qf.argtypes = [py_object, c_int, c_int]
pydll.c_stat_qf.restype=py_object

#----

pydll.c_stat_dgamma.argtypes = [py_object, c_double, c_double]
pydll.c_stat_dgamma.restype=py_object

pydll.c_stat_pgamma.argtypes = [py_object, c_double, c_double]
pydll.c_stat_pgamma.restype=py_object

pydll.c_stat_qgamma.argtypes = [py_object, c_double, c_double]
pydll.c_stat_qgamma.restype=py_object

#----

pydll.c_stat_dgeom.argtypes = [py_object, c_double]
pydll.c_stat_dgeom.restype=py_object

pydll.c_stat_pgeom.argtypes =  [py_object, c_double]
pydll.c_stat_pgeom.restype=py_object

pydll.c_stat_qgeom.argtypes =  [py_object, c_double]
pydll.c_stat_qgeom.restype=py_object

#----

pydll.c_stat_dhyper.argtypes = [py_object, c_int, c_int, c_int]
pydll.c_stat_dhyper.restype=py_object

pydll.c_stat_phyper.argtypes = [py_object, c_int, c_int, c_int]
pydll.c_stat_phyper.restype=py_object

pydll.c_stat_qhyper.argtypes = [py_object, c_int, c_int, c_int]
pydll.c_stat_qhyper.restype=py_object

#----

pydll.c_stat_dnorm.argtypes = [py_object, c_double, c_double]
pydll.c_stat_dnorm.restype=py_object

pydll.c_stat_pnorm.argtypes = [py_object, c_double, c_double]
pydll.c_stat_pnorm.restype=py_object

pydll.c_stat_qnorm.argtypes = [py_object, c_double, c_double]
pydll.c_stat_qnorm.restype=py_object

#----

pydll.c_stat_dlnorm.argtypes = [py_object, c_double, c_double]
pydll.c_stat_dlnorm.restype=py_object

pydll.c_stat_plnorm.argtypes = [py_object, c_double, c_double]
pydll.c_stat_plnorm.restype=py_object

pydll.c_stat_qlnorm.argtypes = [py_object, c_double, c_double]
pydll.c_stat_qlnorm.restype=py_object

#----

pydll.c_stat_dpareto.argtypes = [py_object, c_double, c_double]
pydll.c_stat_dpareto.restype=py_object

pydll.c_stat_ppareto.argtypes = [py_object, c_double, c_double]
pydll.c_stat_ppareto.restype=py_object

pydll.c_stat_qpareto.argtypes = [py_object, c_double, c_double]
pydll.c_stat_qpareto.restype=py_object

#----

pydll.c_stat_dpois.argtypes = [py_object, c_double]
pydll.c_stat_dpois.restype=py_object

pydll.c_stat_ppois.argtypes = [py_object, c_double]
pydll.c_stat_ppois.restype=py_object

pydll.c_stat_qpois.argtypes = [py_object, c_double]
pydll.c_stat_qpois.restype=py_object


#---

pydll.c_stat_psmirnov.argtypes = [py_object, c_int]
pydll.c_stat_psmirnov.restype=py_object


#----

pydll.c_stat_dt.argtypes = [py_object, c_int]
pydll.c_stat_dt.restype=py_object

pydll.c_stat_pt.argtypes = [py_object, c_int]
pydll.c_stat_pt.restype=py_object

pydll.c_stat_qt.argtypes = [py_object, c_int]
pydll.c_stat_qt.restype=py_object

#----

pydll.c_stat_dunif.argtypes = [py_object, c_double, c_double]
pydll.c_stat_dunif.restype=py_object

pydll.c_stat_punif.argtypes = [py_object, c_double, c_double]
pydll.c_stat_punif.restype=py_object

pydll.c_stat_qunif.argtypes = [py_object, c_double, c_double]
pydll.c_stat_qunif.restype=py_object

#----

pydll.c_stat_dweibull.argtypes = [py_object, c_double, c_double]
pydll.c_stat_dweibull.restype=py_object

pydll.c_stat_pweibull.argtypes = [py_object, c_double, c_double]
pydll.c_stat_pweibull.restype=py_object

pydll.c_stat_qweibull.argtypes = [py_object, c_double, c_double]
pydll.c_stat_qweibull.restype=py_object

#----

pydll.c_stat_dsignrank.argtypes = [py_object, c_int]
pydll.c_stat_dsignrank.restype=py_object

pydll.c_stat_psignrank.argtypes = [py_object, c_int]
pydll.c_stat_psignrank.restype=py_object

pydll.c_stat_qsignrank.argtypes = [py_object, c_int]
pydll.c_stat_qsignrank.restype=py_object

#----


pydll.c_stat_rolling.argtypes = [py_object, py_object, c_int]
pydll.c_stat_rolling.restype=py_object

pydll.c_stat_test_norm_ad.argtypes = [py_object]
pydll.c_stat_test_norm_ad.restype=py_object


pydll.c_stat_test_shapirowilkinson.argtypes = [py_object]
pydll.c_stat_test_shapirowilkinson.restype=py_object




"""
--------------------------------------------------------------------------------
----------------------------  PLOT LIBRARY --------------------------------------
"""


pydll.c_plot_boxplot.argtypes = [py_object, py_object]
pydll.c_plot_boxplot.restype=py_object


pydll.c_plot_histogram.argtypes = [py_object, py_object]
pydll.c_plot_histogram.restype=py_object


pydll.c_plot_psychrometry.argtypes = [py_object, py_object]
pydll.c_plot_psychrometry.restype=py_object


pydll.c_plot_scatter.argtypes = [py_object, py_object]
pydll.c_plot_scatter.restype=py_object


pydll.c_plot_canvas.argtypes = [
						py_object, #x-bounds
						py_object, #y-bounds
						c_bool, # Show Horiz Axis
						c_bool, # Show Vert Axis
						c_bool, # Show Horiz Gridlines
						c_bool] # Show Vert Gridlines
pydll.c_plot_canvas.restype=py_object



#-----------------------------------------
pydll.c_plot_gdi_arc.argtypes = [
					c_double, #x1
					c_double, #y1
					c_double, #x2
					c_double, #y2
					c_double, #xcenter
					c_double, #ycenter
					py_object, #dictionary for Pen
					py_object] #dictionary for Brush				
pydll.c_plot_gdi_arc.restype=c_ulonglong


pydll.c_plot_gdi_arrow.argtypes = [
					c_double, #x1
					c_double, #y1
					c_double, #x2
					c_double, #y2
					c_double, #angle
					c_double, #length
					c_char_p, # label
					py_object] #dictionary for Pen				
pydll.c_plot_gdi_arrow.restype=c_ulonglong


pydll.c_plot_gdi_line.argtypes = [
					c_double, #x1
					c_double, #y1
					c_double, #x2
					c_double, #y2
					c_char_p, # label
					py_object] #dictionary for Pen				
pydll.c_plot_gdi_line.restype=c_ulonglong


pydll.c_plot_gdi_rect.argtypes = [
					c_double, #x of top-left
					c_double, #y of top-left
					c_double, #width
					c_double, #height
					c_char_p, # label
					py_object, #dictionary for Pen
					py_object] #dictionary for Brush
pydll.c_plot_gdi_ellipse.restype=c_ulonglong


pydll.c_plot_gdi_ellipse.argtypes = [
					c_double, # x of center
					c_double, # y of center
					c_double, # half width
					c_double, # half height
					c_char_p, # label
					py_object, #dictionary for Pen
					py_object] #dictionary for Brush
pydll.c_plot_gdi_ellipse.restype=c_ulonglong


pydll.c_plot_gdi_text.argtypes = [
					c_double, # x of top-left
					c_double, # y of top-left
					c_char_p, # text
					c_double, # rotation
					c_char, #halign
					c_char, #valign
					c_char_p, # color
					py_object] #dictionary for Font
pydll.c_plot_gdi_text.restype=c_ulonglong


pydll.c_plot_gdi_curve.argtypes = [
					py_object, #x-values (iterable)
					py_object,#y-values (iterable)
					c_char_p, # label
					py_object] #dictionary for Pen
pydll.c_plot_gdi_curve.restype=c_ulonglong


pydll.c_plot_gdi_polygon.argtypes = [
					py_object, #x-values (iterable)
					py_object,#y-values (iterable)
					c_char_p, # label
					py_object, #dictionary for Pen
					py_object] #dictionary for brush
pydll.c_plot_gdi_polygon.restype=c_ulonglong


pydll.c_plot_gdi_marker.argtypes = [
					c_double, # x of centroid
					c_double, # y of centroid
					c_char_p, # type
					c_uint8, # size
					c_char_p, # label
					py_object, # dictionary for Pen
					py_object] #dictionary for Brush
pydll.c_plot_gdi_marker.restype=c_ulonglong


pydll.c_plot_gdi_makegroup.argtypes = [
					c_ulonglong, #owner id
					py_object] #member ids
pydll.c_plot_gdi_makegroup.restype=None


#---------------------------------------------

pydll.c_plot_layout.argtypes = [c_int, c_int]
pydll.c_plot_layout.restype=None


pydll.c_plot_subplot.argtypes = [
						c_int, 
						c_int, 
						c_int, 
						c_int]
pydll.c_plot_subplot.restype=None


pydll.c_plot_figure.argtypes = []
pydll.c_plot_figure.restype=None


pydll.c_plot_title.argtypes = [c_char_p]
pydll.c_plot_title.restype=None


pydll.c_plot_xlabel.argtypes = [c_char_p]
pydll.c_plot_xlabel.restype=None


pydll.c_plot_ylabel.argtypes = [c_char_p]
pydll.c_plot_ylabel.restype=None


pydll.c_plot_axislim.argtypes = [
						py_object, 
						py_object, 
						c_char]
pydll.c_plot_axislim.restype=py_object


pydll.c_plot_set_xticks.argtypes = [
						py_object, 
						py_object,
						c_char_p, #Alignment
						c_char_p] #Position
pydll.c_plot_set_xticks.restype=py_object


pydll.c_plot_set_yticks.argtypes = [
						py_object, 
						py_object,
						c_char_p, #Alignment
						c_char_p] #Position
pydll.c_plot_set_yticks.restype=py_object

#param2: x, y
pydll.c_plot_set_axispos.argtypes = [c_double, c_char]
pydll.c_plot_set_axispos.restype=py_object

#param1: linear, log | param2: x, y
pydll.c_plot_axisscale.argtypes = [c_char_p, c_char]
pydll.c_plot_axisscale.restype=py_object


pydll.c_plot_legend.argtypes = [py_object, py_object]
pydll.c_plot_legend.restype=None


pydll.c_plot_show.argtypes = []
pydll.c_plot_show.restype=None





import ctypes as _ct
from numbers import Real as _Real
from typing import Iterable as _Iterable

from .._ctypeslib import pydll as _pydll
from ._gdiobj import Brush, Font, Pen


def text(
		xy:tuple|list, 
		label:str,
		rotation:float = 0.0,
		hanchor:str = "l",
		vanchor:str = "t",
		**kwargs)->int:
	"""
	xy: (x, y), top-left,
	label: text to be drawn,
	rotation: rotation in degrees (>0 is counter-clockwise)
	hanchor: horizontal anchor, "l", "c" "r" for left, center, right.
	vanchor: vertical anchor, "t", "c", "b" for top, center and bottom
	labelcolor: label color
	"""
	assert isinstance(xy, tuple|list), "xy must be tuple|list."
	assert isinstance(label, str), "label must be string."
	assert isinstance(rotation, _Real), "rotation must be real number."

	_p1 = [i for i in xy if isinstance(i, _Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers."
	
	_color = kwargs.get("labelcolor") or "#000000"
	assert isinstance(_color, str|tuple|list), "color must be str|tuple|list."

	assert isinstance(hanchor, str), "hanchor must be str."
	assert hanchor.upper() in ["L", "C", "R"], "hanchor must be 'l', 'c', 'r' ."

	assert isinstance(vanchor, str), "vanchor must be str"
	assert vanchor.upper() in ["T", "C", "B"], "vanchor must be 't', 'c', 'b' ."

	return _pydll.c_plot_gdi_text(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_char_p(label.encode()),
			_ct.c_double(rotation),
			_ct.c_char(hanchor.upper().encode()),
			_ct.c_char(vanchor.upper().encode()),
			_ct.c_char_p(_color.encode()),
			dict(Font(kwargs)))



def marker(
		xy:tuple|list, 
		type:str = "c",
		size:int = 5,
		label:str = "",
		**kwargs)->int:
	"""
	`xy:`	(x, y), centroid,
	`type:`	type of the marker, "c", "t", "r", "x",
	`size:`	size of the marker in pixels
	"""

	assert isinstance(xy, tuple|list), "xy must be tuple|list."
	assert isinstance(type, str), "type must be string."
	assert isinstance(size, int), "size must be int."
	assert isinstance(label, str), "label must be str."

	assert 1<size<=100, "1 < size <= 100 expected."

	_p1 = [i for i in xy if isinstance(i, _Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers."

	return _pydll.c_plot_gdi_marker(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_char_p(type.encode()),
			_ct.c_uint8(size),
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))



def arc(
		center:tuple|list, 
		p1:tuple|list, 
		p2:tuple|list,
		label:str = "",
		**kwargs)->int:
	"""
	`center:` (x, y) -> center point of arc
	`p1:` (x1, y1) -> start of arc
	`p2:` (x2, y2) -> end of arc

	## Note:
	- The arc is drawn in a counter-clockwise direction between the start and the end points.
	- To plot a circular arc, the plot area must be a square.
	"""


	assert isinstance(center, tuple|list), "center must be tuple|list."
	assert isinstance(p1, tuple|list), "p1 must be tuple|list."
	assert isinstance(p2, tuple|list), "p2 must be tuple|list."
	assert isinstance(label, str), "label must be str."

	_c = [i for i in center if isinstance(i, _Real)]
	assert len(_c) == 2, "center must contain exactly two real numbers."

	_p1 = [i for i in p1 if isinstance(i, _Real)]
	assert len(_p1) == 2, "p1 must contain exactly two real numbers."

	_p2 = [i for i in p2 if isinstance(i, _Real)]
	assert len(_p2) == 2, "p2 must contain exactly two real numbers."

	return _pydll.c_plot_gdi_arc(
			_ct.c_double(p1[0]),
			_ct.c_double(p1[1]),
			_ct.c_double(p2[0]),
			_ct.c_double(p2[1]),
			_ct.c_double(center[0]),
			_ct.c_double(center[1]),
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))



def arrow(
		p1:tuple|list, 
		p2:tuple|list, 
		angle:_Real = 45, #45 degrees
		length:float = 0.1, #10% length of main-line
		label:str = "",
		**kwargs)->int:
	"""
	`p1, p2:` (x1, y1), (x2, y2) coordinate of the main-line
	`angle:` angle between the two head-lines
	`length:` ratio of the length of the head-line to the main-line
	"""

	assert isinstance(p1, tuple|list), "p1 must be tuple|list."
	assert isinstance(p2, tuple|list), "p2 must be tuple|list."
	assert isinstance(angle, _Real), "angle must be Real."
	assert isinstance(length, float), "length must be float."
	assert isinstance(label, str), "label must be str."

	_p1 = [i for i in p1 if isinstance(i, _Real)]
	assert len(_p1) == 2, "p1 must contain exactly two real numbers."

	_p2 = [i for i in p2 if isinstance(i, _Real)]
	assert len(_p2) == 2, "p2 must contain exactly two real numbers."

	assert 5 < angle <=180, "5 < angle <=180 expected." 
	assert 0.01 < length < 1, "0.01 < length < 1 expected."

	return _pydll.c_plot_gdi_arrow(
			_ct.c_double(p1[0]),
			_ct.c_double(p1[1]),
			_ct.c_double(p2[0]),
			_ct.c_double(p2[1]),
			_ct.c_double(angle),
			_ct.c_double(length),
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)))



def curve(
		x: _Iterable, 
		y:_Iterable, 
		label:str = "",
		**kwargs)->int:
	"""
	Draws a smooth curve between (x1, y1), (x2, y2), ..., (xn, yn). 
	The curve is only guaranteed to pass from (x1, y1) and (xn, yn).

	`x:` x values
	`y:` y values
	"""

	assert isinstance(x, _Iterable), "x must be Iterable."
	assert isinstance(y, _Iterable), "y must be Iterable."
	assert isinstance(label, str), "label must be str."

	#pre-check
	assert len(x) == len(y), "x and y must have same lengths."

	_x = [i for i in x if isinstance(i, _Real)]
	assert len(_x) >= 3, "x must contain at least 3 real numbers."

	_y = [i for i in y if isinstance(i, _Real)]
	assert len(_y) >= 3, "y must contain at least 3 real numbers."

	#processed-check
	assert len(_x) == len(_y), "x and y must have same lengths."

	return _pydll.c_plot_gdi_curve(
			x, 
			y, 
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)))



def ellipse(
		xy:tuple|list, 
		width:_Real, 
		height:_Real, 
		label:str = "",
		**kwargs)->int:
	"""
	xy:	 	(x, y), center,
	width: 	half width (>0),
	height: half height (>0),
	"""

	assert isinstance(xy, tuple|list), "p must be tuple|list."
	assert isinstance(width, _Real), "width must be real number."
	assert isinstance(height, _Real), "height must be real number."
	assert isinstance(label, str), "label must be str."
	
	assert width>0, "width>0 expected."
	assert height>0, "height>0 expected."

	_p1 = [i for i in xy if isinstance(i, _Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers."

	if kwargs.get("hatch") == None:
		kwargs["hatch"] = "none"

	return _pydll.c_plot_gdi_ellipse(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_double(width),
			_ct.c_double(height),
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))	



def line(
		p1:tuple|list, 
		p2:tuple|list, 
		label:str = "",
		**kwargs)->int:
	"""
	`p1:` (x1, y1)
	`p2:` (x2, y2)
	`label:` After stripping, if len(label)>0, it is shown in legend
	"""
	assert isinstance(p1, tuple|list), "p1 must be tuple|list."
	assert isinstance(p2, tuple|list), "p2 must be tuple|list."
	assert isinstance(label, str), "label must be str."

	_p1 = [i for i in p1 if isinstance(i, _Real)]
	assert len(_p1) == 2, "p1 must contain exactly two real numbers."

	_p2 = [i for i in p2 if isinstance(i, _Real)]
	assert len(_p2) == 2, "p2 must contain exactly two real numbers."

	return _pydll.c_plot_gdi_line(
			_ct.c_double(p1[0]),
			_ct.c_double(p1[1]),
			_ct.c_double(p2[0]),
			_ct.c_double(p2[1]),
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)))
	


def polygon(
		xy:_Iterable, 
		label:str = "",
		**kwargs)->int:
	"""
	Draws a polygon between (x1, y1), (x2, y2), ..., (xn, yn). 
	The first and last points are automatically closed.

	At least 3 points expected.
	"""

	assert isinstance(xy, _Iterable), "xy must be Iterable."
	assert len(xy)>=3, "x must contain at least 3 Iterables."

	assert isinstance(label, str), "label must be str."
	
	for v in xy:
		assert isinstance(v, _Iterable), "xy must contain Iterables."
		assert len(v) == 2, "Each iterable's length in xy must be exactly equal to 2."
		_x = [i for i in v if isinstance(i, _Real)]
		assert len(_x) == 2, "Each iterable in xy must contain exactly two Real numbers."
	
	x, y = list(zip(*xy))

	return _pydll.c_plot_gdi_polygon(
			x, y, 
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)), 
			dict(Brush(kwargs)))




def rect(
		xy:tuple|list, 
		width:_Real, 
		height:_Real, 
		label:str = "",
		**kwargs)->int:
	"""
	xy:	(x, y), bottom-left corner of the rectangle,
	width: 	width of rectangle (>0),
	height: height of rectangle (>0), 
	"""	

	assert isinstance(xy, tuple|list), "xy must be tuple|list."
	assert isinstance(width, _Real), "width must be real number."
	assert isinstance(height, _Real), "height must be real number."
	assert isinstance(label, str), "label must be str."

	assert width>0, "width>0 expected."
	assert height>0, "height>0 expected."

	_p1 = [i for i in xy if isinstance(i, _Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers."

	kwargs["hatch"] = kwargs.get("hatch") or "none" #make it transparent by default

	return _pydll.c_plot_gdi_rect(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_double(width),
			_ct.c_double(height),
			_ct.c_char_p(label.strip().encode()),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))



def makegroup(
		owner:int, 
		members:_Iterable[int])->None:
	"""
	sets the properties of members based on owner's properties
	
	## Inputs:
	`owner:` A gdi object whose properties will affect members' properties,
	`members:` gdi object(s) whose properties will be synched with owner

	## Note: 
	This function is rather useful (meaningful) if properties will be manipulated at runtime.
	"""
	assert isinstance(owner, int), "owner must be int."

	assert isinstance(members, _Iterable), "members must be Iterable object."
	_target = [i for i in members if isinstance(i, int)]
	assert len(_target) == len(members), "members must contain only int."

	_pydll.c_plot_gdi_makegroup(
		_ct.c_ulonglong(owner), 
		_ct.py_object(set(members)) #unique ids
		)


from ._core import kurt, mode, moveavg, rolling, skew

from ._aov import aov, aov2, aov2_results
from ._linregress import linregress

from ._test_basic import \
	test_f, test_f_Result, \
	test_sign, test_sign_Result,\
	test_t, test_t1_result, test_t2_result, test_tpaired_result, \
	test_z, test_z_Result


from ._test_normality import \
	anderson, ADTestRes, \
	ks_1samp, Ks1SampletestResult, \
	shapiro, ShapiroTestResult


from ._distributions import \
	dbeta, pbeta, qbeta, rbeta, \
	dbinom, pbinom, qbinom, rbinom, \
	dchisq, pchisq, qchisq, rchisq, \
	dexp, pexp, qexp, rexp, \
	df, pf, qf, rf, \
	dgamma, pgamma, qgamma,rgamma, \
	dgeom, pgeom, qgeom, rgeom, \
	dhyper, phyper, qhyper, rhyper, \
	dmultinom, rmultinom, \
	dnbinom, pnbinom, qnbinom,rnbinom, \
	dnorm, pnorm, qnorm, rnorm, \
	dlnorm, plnorm, qlnorm, rlnorm, \
	dpareto, ppareto, qpareto, rpareto, \
	dpois, ppois, qpois, rpois, \
	dsignrank, qsignrank, psignrank, \
	psmirnov, \
	dt, pt, qt, rt, \
	dunif, punif, qunif, runif, \
	dweibull, pweibull, qweibull, rweibull







import numbers
from dataclasses import dataclass
from typing import Iterable
from types import FunctionType

import numpy as _np

from .._ctypeslib import pydll as _pydll
from ._distributions import pnorm, psmirnov





# ----  Anderson-Darling Test -----

@dataclass
class ADTestRes:
	pvalue:float
	A2:float #test statistics


def anderson(x:Iterable)->ADTestRes:
	"""
	Performs Anderson-Darling test
	"""
	assert isinstance(x, Iterable), "x must be an Iterable object"
	
	_xx = [v for v in x if isinstance(v, numbers.Real)]
	assert len(x) == len(_xx), "x must contain only Real numbers"
	
	pval, A2 = _pydll.c_stat_test_norm_ad(x)
	return ADTestRes(pvalue=pval, A2=A2)




# ------- Kolmogorov-Smirnov Test

@dataclass
class Ks1SampletestResult:
	pvalue:float
	D:float #test statistics
	D_loc:float #location of max distance (D)
	D_sign:int
	


def ks_1samp(
		x:Iterable, 
		cdf:FunctionType=pnorm, 
		args:tuple=())->Ks1SampletestResult:
	"""
	Performs two.sided Kolmogorov-Smirnov test

	Reference:
	- Simard R & L'Ecuyer P (2011) "Computing the Two-Sided Kolmogorov-Smirnov Distribution". 
	  J of Statistical Software, 39:11.
	"""
	assert isinstance(x, Iterable), "x must be Iterable"
	assert isinstance(cdf, FunctionType), "cdf must be a function"

	_xx = [v for v in x if isinstance(v, numbers.Real)]
	assert len(x) == len(_xx), "x must contain only Real numbers"

	n = len(x)
	x = _np.sort(x)
	cdfvals = cdf(x, *args)
	
	dplus = (_np.arange(1.0, n + 1) / n - cdfvals)
	_plus = dplus.argmax()

	dminus = (cdfvals - _np.arange(0.0, n)/n)
	_minus = dminus.argmax()

	Dminus, dminus_loc = float(dminus[_minus]), float(x[_minus])
	Dplus, dplus_loc = float(dplus[_plus]), float(x[_plus])
	
	Dvalue = max(Dplus, Dminus)

	return Ks1SampletestResult(
					pvalue = 1-psmirnov(Dvalue, n),
					D = Dvalue, 
					D_sign = 1 if Dplus>Dminus else -1,
					D_loc = dplus_loc if Dplus>Dminus else dminus_loc)




# ----  Shapiro-Wilkinson Test -----

@dataclass
class ShapiroTestResult:
	pvalue:float
	W:float #test statistics
	msg:str #warning message, if exists


def shapiro(x:Iterable)->ShapiroTestResult:
	"""
	Performs Shapiro-Wilkinson test

	- x must be iterable containing only Real numbers
	- x must have at least length 3.
	- if len(x)>5000, W is accurate, but the p-value may not be.

	Reference:
	- Royston P (1995). Remark AS R94: A Remark on Algorithm AS 181: The W-test for Normality
	  J of the Royal Statistical Society. Series C 44:4 pp. 547-551
	- Shapiro SS & Wilk MB (1965). An Analysis of Variance Test for Normality
	  Biometrika Vol. 52, No. 3/4, pp. 591-611
	"""
	assert isinstance(x, Iterable), "x must be an Iterable object."

	N = len(x)
	if N < 3:
		raise ValueError("x must be at least length 3.")
	
	_xx = [v for v in x if isinstance(v, numbers.Real)]
	assert N == len(_xx), "x must contain only Real numbers."
	
	result = _pydll.c_stat_test_shapirowilkinson(x)
	return ShapiroTestResult(W=result[0], pvalue=result[1], msg=result[2])



import ctypes as _ct
import math as _math
import numbers as _numbers
import types as _types
from typing import Iterable

import numpy as _np

from ._ctypeslib import pydll as _pydll




def trapz(
		x:list|_np.ndarray, 
		y:list|_np.ndarray)->float:
    """Uses Numpy's trapz method."""
    return _np.trapz(x=x, y=y)



#------------------------------------------------

def cumtrapz(
		x:list|_np.ndarray, 
		y:list|_np.ndarray)->list:
	"""Computes the left-tailed cumulative area"""
	val = 0.0
	retList =[0.0]

	for i in range(len(x)-1):
		a, b = x[i], x[i + 1]

		if _math.isclose(b, a, abs_tol = 1E-5) or b < a:
			raise ValueError("|X(i+1)-X(i)|>1E-5 expected.")

		temp = (y[i] + y[i + 1])/2.0
		val += (b - a) * temp

		retList.append(val)
	
	return retList



#------------------------------------------------

def simpson(
		x:Iterable, 
		y:Iterable)->float:
	"""Uses simpson's method (1/3 and 3/8)"""

	return _pydll.c_integ_simpson(x, y)



#------------------------------------------------

def romberg(f:_types.FunctionType, 
	a:float, 
	b:float, 
	tol=1E-5, 
	maxiter=100)->float:
	"""
	Romberg integration

	#Input:
	f: Univariate function 
	a, b: limits of integration (b>a) 
	tol: tolerance for error 
	maxiter: Maximum number of iterations
	"""
	assert callable(f), "f must be function"
	assert isinstance(a, _numbers.Real), "a must be real number"
	assert isinstance(b, _numbers.Real), "b must be real number"
	assert isinstance(tol, _numbers.Real), "tol must be real number"
	assert isinstance(maxiter, _numbers.Integral), "maxiter must be int"

	return _pydll.c_integ_romberg(
		f, 
		_ct.c_double(a), 
		_ct.c_double(b), 
		_ct.c_double(tol), 
		_ct.c_int(maxiter))



#------------------------------------------------

def fixed_quad(f:_types.FunctionType, 
	a:float, 
	b:float, 
	n:int=5)->float:
	"""
	Computes definite integral using fixed-order Gaussian quadrature.

	#Input:
	f: Univariate function \n
	a, b: limits of integration (b>a) \n
	n: Order of quadrature integration (2<=n<=6)
	"""
	assert callable(f), "f must be function"
	assert isinstance(a, _numbers.Real), "a must be real number"
	assert isinstance(b, _numbers.Real), "b must be real number"
	assert isinstance(n, _numbers.Integral), "n must be int"

	return _pydll.c_integ_fixed_quad(
		f, 
		_ct.c_double(a), 
		_ct.c_double(b),  
		_ct.c_int(n))


from ._charts import layout, subplot, figure, legend, show, title, \
	 				xlabel, ylabel, xlim, ylim, set_xticks, set_yticks, set_xposition, set_yposition, \
					xscale, yscale

from ._charts import canvas, boxplot, hist, psychrometry, scatter 

from ._quiverchart import quiver, dirfield
from ._qqcharts import qqnorm, qqplot
from ._bubblechart import bubble
from ._barcharts import bar, barh
from ._moody import moody

from ._chartelems import Marker
